"""
Probability density functions compliant with the lmfit package
Author: Cyril Grima <cyril.grima@gmail.com>
"""

import math
import numpy as np
from scipy import stats, integrate
from scipy.special import jv, kv, j0, i0, digamma



def gamma(params, x, data=None, eps=None):
    """Gamma PDF from scipy with adequate variables
    """
    # Initialisation
    mu = params['mu']
    # Debug inputs
    if hasattr(mu, 'value'): mu = mu.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element
    # Model function
    model = stats.gamma.pdf(x, mu, scale = 1)
    model = np.nan_to_num(model)

    if data is None:
        return model
    if eps is None:
        return (model - data) #residual
    return (model - data)/eps



def rayleigh(params, x, data=None, eps=None):
    """Rayleigh PDF from scipy with adequate variables
    """
    # Initialisation
    s = params['s']
    # Debug inputs
    if hasattr(s, 'value'): s = s.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element
    # Model function
    model = stats.rayleigh.pdf(x, scale = s)
    model = np.nan_to_num(model)

    if data is None:
        return model
    if eps is None:
        return (model - data) #residual
    return (model - data)/eps



def rice(params, x, data=None, eps=None, method = None):
    """Rice PDF from scipy with adequate variables
    """
    # Initialisation
    a = params['a']
    s = params['s']
    # Debug inputs
    if hasattr(a, 'value'): a = a.value #debug due to lmfit.minimize
    if hasattr(s, 'value'): s = s.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element
    # Model function
    model = stats.rice.pdf(x, a/s, scale = s)
    #model = (x/s**2) * np.exp(-(x**2+a**2)/(2*s**2)) * np.i0(a*x/s**2)
    #model = np.nan_to_num(model)

    if data is None:
        return model
    if eps is None:
        return (model - data) #residual
    return (model - data)/eps



def k(params, x, data=None, eps=None):
    """K PDF
    """
    # Initialisation
    s = params['s']
    mu = params['mu']
    # Debug inputs
    if hasattr(s, 'value'): s = s.value #idem
    if hasattr(mu, 'value'): mu = mu.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element

    # Model function
    b = np.sqrt(2*mu)/s
    model = 2*(x/2.)**mu *b**(mu+1.) /math.gamma(mu) *kv(mu-1,b*x)
    model = np.nan_to_num(model)

    if data is None:
        return model
    if eps is None:
        return (model - data) #residual
    return (model - data)/eps



def hk(params, x, data=None, eps=None, method = 'analytic'):
    """Homodyne K-distribution from various methods

    Arguments
    ---------
    params : dict
        params for the hk function {'a', 's', 'mu'}
    x : sequence
        x coordinates

    Keywords
    --------
    data : sequence
        data to compare the result with (for minimization)
    eps : sequence
        error on the data
    method : string
        'analytic' = from the common analytic form
        'compound' = from the compound representation [Destrempes and Cloutier,
                     2010, Ultrasound in Med. and Biol. 36(7): 1037-51, Eq. 16]
        NB: 'compound' is less unstable than 'analytic' but is ~10x longer!
    verbose : bool
        print fitting report
    """
    # Initialisation
    a = params['a']
    s = params['s']
    mu = params['mu']
    # Debug inputs
    if hasattr(a, 'value'): a = a.value #debug due to lmfit.minimize
    if hasattr(s, 'value'): s = s.value #idem
    if hasattr(mu, 'value'): mu = mu.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element

    def integrand_analytic(w, x, a, s, mu):
        return x*w*j0(w*a)*j0(w*x)*(1. +w*w*s*s/(2.*mu))**-mu
    def integrand_compound(w, x, a, s, mu):
        return rice({'a':a,'s':s*np.sqrt(w/mu)}, x) * gamma({'mu':mu}, w)

    integrands = {
        'analytic': integrand_analytic,
        'compound': integrand_compound,
    }
    integrand = integrands[method]

    model = [integrate.quad(integrand, 0., np.inf, args=(i, a, s, mu), full_output=1)[0]
            for i in x] # Integration
    model = np.array(model)

    #def integrand(w, x, a, s, mu, method=method):
    #    if method == 'analytic':
    #        return x*w*j0(w*a)*j0(w*x)*(1. +w*w*s*s/(2.*mu))**-mu
    #    if method == 'compound':
    #        return rice({'a':a,'s':s*np.sqrt(w/mu)}, x) * gamma({'mu':mu}, w)

    #model = [integrate.quad(integrand, 0., np.inf, args=(i, a, s, mu, method), full_output=1)[0]
    #         for i in x] # Integration
    #model = np.array(model)

    if data is None:
        return model
    if eps is None:
        return (model - data) #residual
    return (model - data)/eps



"""Various python classes for rsr package
"""

import numpy as np
from . import invert
import matplotlib.pyplot as plt
import subradar as sr
import multiprocessing


class Statfit:
    """Class holding statistical fit results
    """
    def __init__(self, sample, func, values, params, chisqr, redchi,
                 elapsed, nfev, message, success, residual, x, n, edges, bins):
        self.sample = sample
        self.func = func
        self.values = values
        self.params = params
        self.chisqr = chisqr
        self.redchi = redchi
        self.elapsed = elapsed
        self.nfev = nfev
        self.message = message
        self.success = success
        self.residual = residual
        self.x = x
        self.n = n
        self.edges = edges
        self.bins = bins


    def power(self, db=True):
        """Total (pt), coherent (pc), and incoherent (pn) components in power
        """
        pt, pc, pn = np.average(self.sample)**2, self.values['a']**2, \
                     2*self.values['s']**2*self.values['mu']
        if db is True:
            pt, pc, pn = 10*np.log10(pt), 10*np.log10(pc), 10*np.log10(pn)
        pt = 0 if self.success is False else pt
        pc = 0 if self.success is False else pc
        pn = 0 if self.success is False else pn
        return {'pt':pt, 'pc':pc, 'pn':pn, 'pc-pn':pc-pn}


    def crl(self, **kwargs):
        """Correlation coefficient between distribution and theoretical fit
        """
        try:
            out = np.corrcoef(self.n, self.n+self.residual)[0,1]
        except:
            out = 0.

        if np.isfinite(out) is False or self.success is False:
            out = 0.
        return out


    def invert(self, frq=60e6, th_max=1e-4, cl_logrange=[5], n=1e4,
        ep_range=[1,100], method='iem', approx='Small_S', ):
        """Invert signal components into physical properties
        """
        if method == 'spm':
            out =  getattr(invert, method)(frq, self.power()['pc'],
                   self.power()['pn'])
        if method == 'iem':
            out = sr.invert.power2srf_norminc(method, approx,
            self.power()['pc'], self.power()['pn'], ep_range=ep_range,
            th_max=th_max, wf=frq, cl_logrange=cl_logrange, n=n)
        return out


    def plot(self, ylabel='Normalized Probability', xlabel='Amplitude',
        color='k', ls='-', bins=None, fbins=100, alpha=.1, 
        method='compound', histtype='stepfilled', xlim=None):
        """Plot histogram and pdf
        """
        if bins is None: bins = self.bins

        foo, edges, bar = plt.hist(self.sample, bins=bins, density=True)
        x = [ val-(val-edges[i-1])/2. for i, val in enumerate(edges) ][1:]

        plt.plot(x, self.func(self.values, x, method=method), color=color,
                 ls=ls, linewidth=2)
        plt.xlim(xlim)
        plt.ylabel(ylabel, size=17)
        plt.xlabel(xlabel, size=17)
        plt.yticks(size='17')
        plt.xticks(size='17')


    def report(self, frq=60e6, inv='spm'):
        """Print a report for the fit
        """
        buff = []
        add = buff.append

        add('[%i] [%6.2f s.] [%3.0f eval.] [%3.3f]\n'
            % (self.flag(), self.elapsed, self.nfev, self.crl()))

        add('%s\n' % (self.message))

        #for i, key in enumerate(self.values.keys()):
        #    add('%s = %.3e, ' % (key, self.params[key].value))

        #add('\n')

        for i, key in enumerate(self.power().keys()):
            add('%s = %3.1f dB, ' % (key, self.power()[key]))

        add('\n')

        p = self.invert(frq=frq, method=inv)
        add('%s @ %.0f MHz gives, eps = %.3f, sh = %.2e m' % (inv.upper(),
            frq*1e-6, p['eps'], p['sh']))

        add('\n')

        out = "".join(buff)
        print(out)


    def flag(self):
        """0 is bad data, 1 is good data
        """
        out = self.success * np.isfinite(self.crl()) * (self.crl() > 0)
        return(int(out))



class Async:
    """
    Class to support multi procesing jobs. For calling example, see:
    http://masnun.com/2014/01/01/async-execution-in-python-using-multiprocessing-pool.html
    """
    def __init__(self, func, cb_func, nbcores=1):
        self.func = func
        self.cb_func = cb_func
        self.pool = multiprocessing.Pool(nbcores)

    def call(self,*args, **kwargs):
        return self.pool.apply_async(self.func, args, kwargs, self.cb_func)

    def wait(self):
        self.pool.close()
        self.pool.join()





import numpy as np
from . import pdf as  funcs
from . import run
import scipy


def inverse_cdf(f, xN=1e3, xlim=(0,1), **kwargs):
    """Provides the inverse cumulative density function (CDF) of a probability 
    density function
    
    ARGUMENTS
    ---------
    f : function
        The PDF to evaluate
    xN : integer
        The x-axis sampling of the PDF
    xlim : (integer, integer)
        The x-axis range overwhich the pdf is considered
        
    RETURN
    ------
    the inverse CDF
    """    
    # Probability Density Function
    x = np.linspace(xlim[0],xlim[1],int(xN))
    #func = f(x)
    # Cumulative Distribution Function
    cdf = np.cumsum(f(x))
    # Normalisation to 1
    cdf = cdf/cdf.max()
    # Inverse CDF
    return scipy.interpolate.interp1d(cdf,x)


def sample(f, N=1e3, xN=1e3, xlim=(0,1), precision=None, **kwargs):
    """Provides a random trial X over a given PDF following the Inverse 
    Transform Sampling technique so that X=F^{-1}(U), where F^{-1} is the 
    inverse cumulative distribution function of the PDF and U is a random 
    variable uniformly distributed between 0 and 1.
    
    ARGUMENTS
    ---------
    f : function
        The PDF to evaluate
    N : Integer
        The number of random evaluation to make
    xN : integer
        The x-axis sampling of the PDF
    xlim : (integer, integer)
        The x-axis range overwhich the pdf is considered
    precision : float
        precision on the measurement [dB]
        
    RETURN
    ------
    N random evaluation over the considered PDF
    """
    uniform_samples = np.random.random(int(N))
    required_samples = inverse_cdf(f, xN=xN, xlim=xlim)(uniform_samples)

    if precision:
        for i, sample in enumerate(required_samples):
            linear_precision = np.abs(10**(precision/2/20) - 
                                      10**(-precision/2/20))*sample
            required_samples[i] = np.random.normal(sample, linear_precision)
    
    return required_samples


def pdf(func, params, method=None, **kwargs):
    """Provides a random trial X over a PDF in rsr.pdf
    
    ARGUMENTS
    ---------
    func : string
        PDF name in rsr.pdf
    params : dict
        parameters to be passed to the PDF function
    kwargs : dict
        Any arguments used bu sample.sample
        
    RETURN
    ------
    N random evaluation over the considered PDF
    """
    if method is None:
        f = lambda x: getattr(funcs, func)(params, x,)
    else:
        f = lambda x: getattr(funcs, func)(params, x, method=method)
    return sample(f, **kwargs)


def rsr(func, params, method=None, **kwargs):
    """Apply the RSR over a set of generated random amplitudes
    
    ARGUMENTS
    ---------
    func : string
        PDF name in rsr.pdf
    params : dict
        parameters to be passed to the PDF function
    kwargs : dict
        Any arguments used bu sample.sample

    RETURN
    ------
    N random evaluation over the considered PDF
    """
    amp = pdf(func, params, method=method, **kwargs)
    p = run.processor(amp)
    return p


def power_to_params(pc_db, pn_db, mu=100):
    """Converts Pc and Pn powers in dB to a params dictionary
    """
    pc, pn = 10**(pc_db/10), 10**(pn_db/10)
    params = {'a':np.sqrt(pc), 
              's':np.sqrt(pn/2.), 
              'mu':mu}
    return params


def params_to_power(params, dB=True):
    """Converts a params dict into pc and pn powers in dB
    """
    pc = params['a']**2
    pn = 2*params['s']**2
    if dB == True:
        pc = 10*np.log10(pc)
        pn = 10*np.log10(pn)
    out = {'pc':pc, 'pn':pn}
    return out


def effective_precision(func, params, Nsets=1, **kwargs):
    """Gets the effective precision applicabale to the derivation of Pc and Pn from the rsr algorithm
    the precision is the median of Nsets of histograms randomly obtained from params.

    ARGUMENTS
    ---------
    func : string
        PDF name in rsr.pdf
    params : dict
        parameters to be passed to the PDF function
    Nsets: integer
        Number of amplitude histogram to draw
    kwargs : dict
        Any arguments used bu sample.sample

    RETURN
    ------
    pc and pn median precisions

   """
    power = params_to_power(params, dB=True)
    ps = [rsr(func, params, **kwargs) for i in np.arange(Nsets)]
    d_pcs = [p.power()['pc']-power['pc'] for p in ps]
    d_pns = [p.power()['pn']-power['pn'] for p in ps]
    out = {'ps':ps, 'power':power, 'd_pcs':d_pcs, 'd_pns':d_pns, 'd_pc':np.median(d_pcs), 'd_pn':np.median(d_pns)}
    return out
